@@ -33,10 +33,12 @@ import sys, os
 class Command(object):
     finish = 0
     kills_digit_arg = 1
-    def __init__(self, reader, (event_name, event)):
+
+    def __init__(self, reader, event_name, event):
         self.reader = reader
         self.event = event
         self.event_name = event_name
+
     def do(self):
         pass
 
@@ -368,7 +370,7 @@ class invalid_command(Command):
 class qIHelp(Command):
     def do(self):
         r = self.reader
-        r.insert((self.event + r.console.getpending().data) * r.get_arg())
+        r.insert((bytes(self.event) + r.console.getpending().data) * r.get_arg())
         r.pop_input_trans()
 
 from pyrepl import input
