@@ -19,64 +19,67 @@
 # CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 # CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
-import types, unicodedata
+from __future__ import unicode_literals
+import unicodedata
 from pyrepl import commands
 from pyrepl import input
+try:
+    unicode
+except NameError:
+    unicode = str
+    unichr = chr
+    basestring = bytes, str
+
 
 def _make_unctrl_map():
     uc_map = {}
     for c in map(unichr, range(256)):
-        if unicodedata.category(c)[0] <> 'C':
+        if unicodedata.category(c)[0] != 'C':
             uc_map[c] = c
     for i in range(32):
         c = unichr(i)
-        uc_map[c] = u'^' + unichr(ord('A') + i - 1)
-    uc_map['\t'] = '    ' # display TABs as 4 characters
-    uc_map['\177'] = u'^?'
+        uc_map[c] = '^' + unichr(ord('A') + i - 1)
+    uc_map[b'\t'] = '    '  # display TABs as 4 characters
+    uc_map[b'\177'] = unicode('^?')
     for i in range(256):
         c = unichr(i)
-        if not uc_map.has_key(c):
-            uc_map[c] = u'\\%03o'%i 
+        if c not in uc_map:
+            uc_map[c] = unicode('\\%03o') % i
     return uc_map
 
-# disp_str proved to be a bottleneck for large inputs, so it's been
-# rewritten in C; it's not required though.
-try:
-    raise ImportError # currently it's borked by the unicode support
-    
-    from _pyrepl_utils import disp_str, init_unctrl_map
-    
-    init_unctrl_map(_make_unctrl_map())
 
-    del init_unctrl_map
-except ImportError:
-    def _my_unctrl(c, u=_make_unctrl_map()):
-        import unicodedata
-        if c in u:
-            return u[c]
+def _my_unctrl(c, u=_make_unctrl_map()):
+    if c in u:
+        return u[c]
+    else:
+        if unicodedata.category(c).startswith('C'):
+            return b'\u%04x' % ord(c)
         else:
-            if unicodedata.category(c).startswith('C'):
-                return '\u%04x'%(ord(c),)
-            else:
-                return c
+            return c
 
-    def disp_str(buffer, join=''.join, uc=_my_unctrl):
-        """ disp_str(buffer:string) -> (string, [int])
 
-        Return the string that should be the printed represenation of
-        |buffer| and a list detailing where the characters of |buffer|
-        get used up.  E.g.:
+def disp_str(buffer, join=''.join, uc=_my_unctrl):
+    """ disp_str(buffer:string) -> (string, [int])
 
-        >>> disp_str(chr(3))
-        ('^C', [1, 0])
+    Return the string that should be the printed represenation of
+    |buffer| and a list detailing where the characters of |buffer|
+    get used up.  E.g.:
 
-        the list always contains 0s or 1s at present; it could conceivably
-        go higher as and when unicode support happens."""
-        s = map(uc, buffer)
-        return (join(s),
-                map(ord, join(map(lambda x:'\001'+(len(x)-1)*'\000', s))))
+    >>> disp_str(chr(3))
+    ('^C', [1, 0])
 
-    del _my_unctrl
+    the list always contains 0s or 1s at present; it could conceivably
+    go higher as and when unicode support happens."""
+    # disp_str proved to be a bottleneck for large inputs,
+    # so it needs to be rewritten in C; it's not required though.
+    s = [uc(x) for x in buffer]
+    b = []  # XXX: bytearray
+    for x in s:
+        b.append(1)
+        b.extend([0] * (len(x) - 1))
+    return join(s), b
+
+del _my_unctrl
 
 del _make_unctrl_map
 
@@ -86,6 +89,7 @@ del _make_unctrl_map
  SYNTAX_WORD,
  SYNTAX_SYMBOL] = range(3)
 
+
 def make_default_syntax_table():
     # XXX perhaps should use some unicodedata here?
     st = {}
@@ -93,7 +97,7 @@ def make_default_syntax_table():
         st[c] = SYNTAX_SYMBOL
     for c in [a for a in map(unichr, range(256)) if a.isalpha()]:
         st[c] = SYNTAX_WORD
-    st[u'\n'] = st[u' '] = SYNTAX_WHITESPACE
+    st[unicode('\n')] = st[unicode(' ')] = SYNTAX_WHITESPACE
     return st
 
 default_keymap = tuple(
@@ -118,7 +122,7 @@ default_keymap = tuple(
      (r'\C-x\C-u', 'upcase-region'),
      (r'\C-y', 'yank'),
      (r'\C-z', 'suspend'),
-     
+
      (r'\M-b', 'backward-word'),
      (r'\M-c', 'capitalize-word'),
      (r'\M-d', 'kill-word'),
@@ -139,11 +143,11 @@ default_keymap = tuple(
      (r'\M-8', 'digit-arg'),
      (r'\M-9', 'digit-arg'),
      #(r'\M-\n', 'insert-nl'),
-     ('\\\\', 'self-insert')] + \
+     ('\\\\', 'self-insert')] +
     [(c, 'self-insert')
-     for c in map(chr, range(32, 127)) if c <> '\\'] + \
+     for c in map(chr, range(32, 127)) if c != '\\'] +
     [(c, 'self-insert')
-     for c in map(chr, range(128, 256)) if c.isalpha()] + \
+     for c in map(chr, range(128, 256)) if c.isalpha()] +
     [(r'\<up>', 'up'),
      (r'\<down>', 'down'),
      (r'\<left>', 'left'),
@@ -152,15 +156,19 @@ default_keymap = tuple(
      (r'\<delete>', 'delete'),
      (r'\<backspace>', 'backspace'),
      (r'\M-\<backspace>', 'backward-kill-word'),
-     (r'\<end>', 'end'),
-     (r'\<home>', 'home'),
+     (r'\<end>', 'end-of-line'),         # was 'end'
+     (r'\<home>', 'beginning-of-line'),  # was 'home'
      (r'\<f1>', 'help'),
-     (r'\EOF', 'end'),  # the entries in the terminfo database for xterms
-     (r'\EOH', 'home'), # seem to be wrong.  this is a less than ideal
-                        # workaround
+     (r'\EOF', 'end'),   # the entries in the terminfo database for xterms
+     (r'\EOH', 'home'),  # seem to be wrong.  this is a less than ideal
+                         # workaround
+     (r'\<ctrl left>',  'backward-word'),
+     (r'\<ctrl right>', 'forward-word'),
      ])
 
-del c # from the listcomps
+if 'c' in globals():  # only on python 2.x
+    del c  # from the listcomps
+
 
 class Reader(object):
     """The Reader class implements the bare bones of a command reader,
@@ -238,9 +246,9 @@ feeling more loquacious than I am now."""
         self.commands = {}
         self.msg = ''
         for v in vars(commands).values():
-            if  ( isinstance(v, type)
-                  and issubclass(v, commands.Command)
-                  and v.__name__[0].islower() ):
+            if (isinstance(v, type) and
+                    issubclass(v, commands.Command) and
+                    v.__name__[0].islower()):
                 self.commands[v.__name__] = v
                 self.commands[v.__name__.replace('_', '-')] = v
         self.syntax_table = make_default_syntax_table()
@@ -274,21 +282,25 @@ feeling more loquacious than I am now."""
                         screeninfo.append((0, []))
                 self.lxy = p, ln
             prompt = self.get_prompt(ln, ll >= p >= 0)
+            while '\n' in prompt:
+                pre_prompt, _, prompt = prompt.partition('\n')
+                screen.append(pre_prompt)
+                screeninfo.append((0, []))
             p -= ll + 1
-            lp = len(prompt)
+            prompt, lp = self.process_prompt(prompt)
             l, l2 = disp_str(line)
-            wrapcount = (len(l) + lp) / w
+            wrapcount = (len(l) + lp) // w
             if wrapcount == 0:
                 screen.append(prompt + l)
-                screeninfo.append((lp, l2+[1]))
+                screeninfo.append((lp, l2 + [1]))
             else:
-                screen.append(prompt + l[:w-lp] + "\\")
-                screeninfo.append((lp, l2[:w-lp]))
-                for i in range(-lp + w, -lp + wrapcount*w, w):
-                    screen.append(l[i:i+w] +  "\\")
+                screen.append(prompt + l[:w - lp] + "\\")
+                screeninfo.append((lp, l2[:w - lp]))
+                for i in range(-lp + w, -lp + wrapcount * w, w):
+                    screen.append(l[i:i + w] + "\\")
                     screeninfo.append((0, l2[i:i + w]))
-                screen.append(l[wrapcount*w - lp:])
-                screeninfo.append((0, l2[wrapcount*w - lp:]+[1]))
+                screen.append(l[wrapcount * w - lp:])
+                screeninfo.append((0, l2[wrapcount * w - lp:] + [1]))
         self.screeninfo = screeninfo
         self.cxy = self.pos2xy(self.pos)
         if self.msg and self.msg_at_bottom:
@@ -297,6 +309,31 @@ feeling more loquacious than I am now."""
                 screeninfo.append((0, []))
         return screen
 
+    def process_prompt(self, prompt):
+        """ Process the prompt.
+
+        This means calculate the length of the prompt. The character \x01
+        and \x02 are used to bracket ANSI control sequences and need to be
+        excluded from the length calculation.  So also a copy of the prompt
+        is returned with these control characters removed.  """
+
+        out_prompt = ''
+        l = len(prompt)
+        pos = 0
+        while True:
+            s = prompt.find('\x01', pos)
+            if s == -1:
+                break
+            e = prompt.find('\x02', s)
+            if e == -1:
+                break
+            # Found start and end brackets, subtract from string length
+            l = l - (e - s + 1)
+            out_prompt += prompt[pos:s] + prompt[s + 1:e]
+            pos = e + 1
+        out_prompt += prompt[pos:]
+        return out_prompt, l
+
     def bow(self, p=None):
         """Return the 0-based index of the word break preceding p most
         immediately.
@@ -308,7 +345,7 @@ feeling more loquacious than I am now."""
         st = self.syntax_table
         b = self.buffer
         p -= 1
-        while p >= 0 and st.get(b[p], SYNTAX_WORD) <> SYNTAX_WORD:
+        while p >= 0 and st.get(b[p], SYNTAX_WORD) != SYNTAX_WORD:
             p -= 1
         while p >= 0 and st.get(b[p], SYNTAX_WORD) == SYNTAX_WORD:
             p -= 1
@@ -324,7 +361,7 @@ feeling more loquacious than I am now."""
             p = self.pos
         st = self.syntax_table
         b = self.buffer
-        while p < len(b) and st.get(b[p], SYNTAX_WORD) <> SYNTAX_WORD:
+        while p < len(b) and st.get(b[p], SYNTAX_WORD) != SYNTAX_WORD:
             p += 1
         while p < len(b) and st.get(b[p], SYNTAX_WORD) == SYNTAX_WORD:
             p += 1
@@ -340,10 +377,10 @@ feeling more loquacious than I am now."""
             p = self.pos
         b = self.buffer
         p -= 1
-        while p >= 0 and b[p] <> '\n':
+        while p >= 0 and b[p] != '\n':
             p -= 1
         return p + 1
-    
+
     def eol(self, p=None):
         """Return the 0-based index of the line break following p most
         immediately.
@@ -352,7 +389,7 @@ feeling more loquacious than I am now."""
         if p is None:
             p = self.pos
         b = self.buffer
-        while p < len(b) and b[p] <> '\n':
+        while p < len(b) and b[p] != '\n':
             p += 1
         return p
 
@@ -369,16 +406,22 @@ feeling more loquacious than I am now."""
         """Return what should be in the left-hand margin for line
         `lineno'."""
         if self.arg is not None and cursor_on_line:
-            return "(arg: %s) "%self.arg
+            return "(arg: %s) " % self.arg
         if "\n" in self.buffer:
             if lineno == 0:
-                return self._ps2
+                res = self.ps2
             elif lineno == self.buffer.count("\n"):
-                return self._ps4
+                res = self.ps4
             else:
-                return self._ps3
+                res = self.ps3
         else:
-            return self._ps1
+            res = self.ps1
+        # Lazily call str() on self.psN, and cache the results using as key
+        # the object on which str() was called.  This ensures that even if the
+        # same object is used e.g. for ps1 and ps2, str() is called only once.
+        if res not in self._pscache:
+            self._pscache[res] = str(res)
+        return self._pscache[res]
 
     def push_input_trans(self, itrans):
         self.input_trans_stack.append(self.input_trans)
@@ -428,7 +471,7 @@ feeling more loquacious than I am now."""
         """This function is called to allow post command cleanup."""
         if getattr(cmd, "kills_digit_arg", 1):
             if self.arg is not None:
-                self.dirty = 1                
+                self.dirty = 1
             self.arg = None
 
     def prepare(self):
@@ -444,8 +487,7 @@ feeling more loquacious than I am now."""
             self.pos = 0
             self.dirty = 1
             self.last_command = None
-            self._ps1, self._ps2, self._ps3, self._ps4 = \
-                           map(str, [self.ps1, self.ps2, self.ps3, self.ps4])
+            self._pscache = {}
         except:
             self.restore()
             raise
@@ -477,15 +519,18 @@ feeling more loquacious than I am now."""
         # this call sets up self.cxy, so call it first.
         screen = self.calc_screen()
         self.console.refresh(screen, self.cxy)
-        self.dirty = 0 # forgot this for a while (blush)
+        self.dirty = 0  # forgot this for a while (blush)
 
     def do_cmd(self, cmd):
         #print cmd
-        if isinstance(cmd[0], str):
+        if isinstance(cmd[0], basestring):
+            #XXX: unify to text
             cmd = self.commands.get(cmd[0],
-                                    commands.invalid_command)(self, cmd)
+                                    commands.invalid_command)(self, *cmd)
         elif isinstance(cmd[0], type):
-            cmd = cmd[0](self, cmd)
+            cmd = cmd[0](self, *cmd)
+        else:
+            return  # nothing to do
 
         cmd.do()
 
@@ -515,9 +560,11 @@ feeling more loquacious than I am now."""
 
         while 1:
             event = self.console.get_event(block)
-            if not event: # can only happen if we're not blocking
+            if not event:  # can only happen if we're not blocking
                 return None
 
+            translate = True
+
             if event.evt == 'key':
                 self.input_trans.push(event)
             elif event.evt == 'scroll':
@@ -525,9 +572,12 @@ feeling more loquacious than I am now."""
             elif event.evt == 'resize':
                 self.refresh()
             else:
-                pass
+                translate = False
 
-            cmd = self.input_trans.get()
+            if translate:
+                cmd = self.input_trans.get()
+            else:
+                cmd = event.evt, event.data
 
             if cmd is None:
                 if block:
@@ -541,16 +591,20 @@ feeling more loquacious than I am now."""
     def push_char(self, char):
         self.console.push_char(char)
         self.handle1(0)
-    
-    def readline(self):
+
+    def readline(self, returns_unicode=False, startup_hook=None):
         """Read a line.  The implementation of this method also shows
         how to drive Reader if you want more control over the event
         loop."""
         self.prepare()
         try:
+            if startup_hook is not None:
+                startup_hook()
             self.refresh()
             while not self.finished:
                 self.handle1()
+            if returns_unicode:
+                return self.get_unicode()
             return self.get_buffer()
         finally:
             self.restore()
@@ -565,11 +619,12 @@ feeling more loquacious than I am now."""
     def get_buffer(self, encoding=None):
         if encoding is None:
             encoding = self.console.encoding
-        return u''.join(self.buffer).encode(self.console.encoding)
+        return unicode('').join(self.buffer).encode(self.console.encoding)
 
     def get_unicode(self):
         """Return the current buffer as a unicode string."""
-        return u''.join(self.buffer)
+        return unicode('').join(self.buffer)
+
 
 def test():
     from pyrepl.unix_console import UnixConsole
@@ -581,5 +636,5 @@ def test():
     while reader.readline():
         pass
 
-if __name__=='__main__':
+if __name__ == '__main__':
     test()
