@@ -19,15 +19,32 @@
 # CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 # CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
-import termios, select, os, struct, errno
-import signal, re, time, sys
+import termios
+import select
+import os
+import struct
+import errno
+import signal
+import re
+import time
+import sys
 from fcntl import ioctl
-from pyrepl import curses
-from pyrepl.fancy_termios import tcgetattr, tcsetattr
-from pyrepl.console import Console, Event
-from pyrepl import unix_eventqueue
+from . import curses
+from .fancy_termios import tcgetattr, tcsetattr
+from .console import Console, Event
+from .unix_eventqueue import EventQueue
+from .trace import trace
 
-_error = (termios.error, curses.error)
+
+class InvalidTerminal(RuntimeError):
+    pass
+
+try:
+    unicode
+except NameError:
+    unicode = str
+
+_error = (termios.error, curses.error, InvalidTerminal)
 
 # there are arguments for changing this to "refresh"
 SIGWINCH_EVENT = 'repaint'
@@ -35,16 +52,18 @@ SIGWINCH_EVENT = 'repaint'
 FIONREAD = getattr(termios, "FIONREAD", None)
 TIOCGWINSZ = getattr(termios, "TIOCGWINSZ", None)
 
+
 def _my_getstr(cap, optional=0):
     r = curses.tigetstr(cap)
     if not optional and r is None:
-        raise RuntimeError, \
-              "terminal doesn't have the required '%s' capability"%cap
+        raise InvalidTerminal(
+            "terminal doesn't have the required '%s' capability" % cap)
     return r
 
+
 # at this point, can we say: AAAAAAAAAAAAAAAAAAAAAARGH!
 def maybe_add_baudrate(dict, rate):
-    name = 'B%d'%rate
+    name = 'B%d' % rate
     if hasattr(termios, name):
         dict[getattr(termios, name)] = rate
 
@@ -55,7 +74,7 @@ for r in [0, 110, 115200, 1200, 134, 150, 1800, 19200, 200, 230400,
 
 del r, maybe_add_baudrate
 
-delayprog = re.compile("\\$<([0-9]+)((?:/|\\*){0,2})>")
+delayprog = re.compile(b"\\$<([0-9]+)((?:/|\\*){0,2})>")
 
 try:
     poll = select.poll
@@ -65,19 +84,28 @@ except AttributeError:
     class poll:
         def __init__(self):
             pass
+
         def register(self, fd, flag):
             self.fd = fd
+
         def poll(self, timeout=None):
-            r,w,e = select.select([self.fd],[],[],timeout)
+            r, w, e = select.select([self.fd], [], [], timeout)
             return r
 
 POLLIN = getattr(select, "POLLIN", None)
 
+
+required_curses_tistrings = 'bel clear cup el'
+optional_curses_tistrings = (
+    'civis cnorm cub cub1 cud cud1 cud cud1 cuf '
+    'cuf1 cuu cuu1 dch dch1 hpa ich ich1 ind pad ri rmkx smkx')
+
+
 class UnixConsole(Console):
     def __init__(self, f_in=0, f_out=1, term=None, encoding=None):
         if encoding is None:
             encoding = sys.getdefaultencoding()
-            
+
         self.encoding = encoding
 
         if isinstance(f_in, int):
@@ -89,53 +117,35 @@ class UnixConsole(Console):
             self.output_fd = f_out
         else:
             self.output_fd = f_out.fileno()
-        
+
         self.pollob = poll()
         self.pollob.register(self.input_fd, POLLIN)
         curses.setupterm(term, self.output_fd)
         self.term = term
-        
-        self._bel   = _my_getstr("bel")
-        self._civis = _my_getstr("civis", optional=1)
-        self._clear = _my_getstr("clear")
-        self._cnorm = _my_getstr("cnorm", optional=1)
-        self._cub   = _my_getstr("cub",   optional=1)
-        self._cub1  = _my_getstr("cub1",  1)
-        self._cud   = _my_getstr("cud",   1)
-        self._cud1  = _my_getstr("cud1",  1)
-        self._cuf   = _my_getstr("cuf",   1)
-        self._cuf1  = _my_getstr("cuf1",  1)
-        self._cup   = _my_getstr("cup")
-        self._cuu   = _my_getstr("cuu",   1)
-        self._cuu1  = _my_getstr("cuu1",  1)
-        self._dch1  = _my_getstr("dch1",  1)
-        self._dch   = _my_getstr("dch",   1)
-        self._el    = _my_getstr("el")
-        self._hpa   = _my_getstr("hpa",   1)
-        self._ich   = _my_getstr("ich",   1)
-        self._ich1  = _my_getstr("ich1",  1)
-        self._ind   = _my_getstr("ind",   1)
-        self._pad   = _my_getstr("pad",   1)
-        self._ri    = _my_getstr("ri",    1)
-        self._rmkx  = _my_getstr("rmkx",  1)
-        self._smkx  = _my_getstr("smkx",  1)
-        
+
+        for name in required_curses_tistrings.split():
+            setattr(self, '_' + name, _my_getstr(name))
+
+        for name in optional_curses_tistrings.split():
+            setattr(self, '_' + name, _my_getstr(name, optional=1))
+
         ## work out how we're going to sling the cursor around
-        if 0 and self._hpa: # hpa don't work in windows telnet :-(
+        # hpa don't work in windows telnet :-(
+        if 0 and self._hpa:
             self.__move_x = self.__move_x_hpa
         elif self._cub and self._cuf:
             self.__move_x = self.__move_x_cub_cuf
         elif self._cub1 and self._cuf1:
             self.__move_x = self.__move_x_cub1_cuf1
         else:
-            raise RuntimeError, "insufficient terminal (horizontal)"
+            raise RuntimeError("insufficient terminal (horizontal)")
 
         if self._cuu and self._cud:
             self.__move_y = self.__move_y_cuu_cud
         elif self._cuu1 and self._cud1:
             self.__move_y = self.__move_y_cuu1_cud1
         else:
-            raise RuntimeError, "insufficient terminal (vertical)"
+            raise RuntimeError("insufficient terminal (vertical)")
 
         if self._dch1:
             self.dch1 = self._dch1
@@ -153,16 +163,12 @@ class UnixConsole(Console):
 
         self.__move = self.__move_short
 
-        self.event_queue = unix_eventqueue.EventQueue(self.input_fd)
-        self.partial_char = ''
+        self.event_queue = EventQueue(self.input_fd, self.encoding)
         self.cursor_visible = 1
 
-    def change_encoding(self, encoding):
-        self.encoding = encoding
-    
-    def refresh(self, screen, (cx, cy)):
+    def refresh(self, screen, c_xy):
         # this function is still too long (over 90 lines)
-
+        cx, cy = c_xy
         if not self.__gone_tall:
             while len(self.screen) < min(len(screen), self.height):
                 self.__hide_cursor()
@@ -172,7 +178,7 @@ class UnixConsole(Console):
                 self.screen.append("")
         else:
             while len(self.screen) < len(screen):
-                self.screen.append("")            
+                self.screen.append("")
 
         if len(screen) > self.height:
             self.__gone_tall = 1
@@ -182,17 +188,6 @@ class UnixConsole(Console):
         old_offset = offset = self.__offset
         height = self.height
 
-        if 0:
-            global counter
-            try:
-                counter
-            except NameError:
-                counter = 0
-            self.__write_code(curses.tigetstr("setaf"), counter)
-            counter += 1
-            if counter > 8:
-                counter = 0
-
         # we make sure the cursor is on the screen, and that we're
         # using all of the screen if we can
         if cy < offset:
@@ -231,7 +226,7 @@ class UnixConsole(Console):
                                         newscr):
             if oldline != newline:
                 self.__write_changed_line(y, oldline, newline, px)
-                
+
         y = len(newscr)
         while y < len(oldscr):
             self.__hide_cursor()
@@ -241,7 +236,7 @@ class UnixConsole(Console):
             y += 1
 
         self.__show_cursor()
-        
+
         self.screen = screen
         self.move_cursor(cx, cy)
         self.flushoutput()
@@ -257,11 +252,12 @@ class UnixConsole(Console):
         # reuse the oldline as much as possible, but stop as soon as we
         # encounter an ESCAPE, because it might be the start of an escape
         # sequene
+        #XXX unicode check!
         while x < minlen and oldline[x] == newline[x] and newline[x] != '\x1b':
             x += 1
         if oldline[x:] == newline[x+1:] and self.ich1:
-            if ( y == self.__posxy[1] and x > self.__posxy[0]
-                 and oldline[px:x] == newline[px+1:x+1] ):
+            if (y == self.__posxy[1] and x > self.__posxy[0] and
+                    oldline[px:x] == newline[px+1:x+1]):
                 x = px
             self.__move(x, y)
             self.__write_code(self.ich1)
@@ -290,10 +286,18 @@ class UnixConsole(Console):
             self.__write(newline[x:])
             self.__posxy = len(newline), y
 
+        #XXX: check for unicode mess
+        if '\x1b' in newline:
+            # ANSI escape characters are present, so we can't assume
+            # anything about the position of the cursor.  Moving the cursor
+            # to the left margin should work to get to a known position.
+            self.move_cursor(0, y)
+
     def __write(self, text):
         self.__buffer.append((text, 0))
 
     def __write_code(self, fmt, *args):
+
         self.__buffer.append((curses.tparm(fmt, *args), 1))
 
     def __maybe_write_code(self, fmt, *args):
@@ -352,13 +356,14 @@ class UnixConsole(Console):
         # per-readline preparations:
         self.__svtermstate = tcgetattr(self.input_fd)
         raw = self.__svtermstate.copy()
-        raw.iflag &=~ (termios.BRKINT | termios.INPCK |
+        raw.iflag |= termios.ICRNL
+        raw.iflag &= ~(termios.BRKINT | termios.INPCK |
                        termios.ISTRIP | termios.IXON)
-        raw.oflag &=~ (termios.OPOST)
-        raw.cflag &=~ (termios.CSIZE|termios.PARENB)
-        raw.cflag |=  (termios.CS8)
-        raw.lflag &=~ (termios.ICANON|termios.ECHO|
-                       termios.IEXTEN|(termios.ISIG*1))
+        raw.oflag &= ~termios.OPOST
+        raw.cflag &= ~(termios.CSIZE | termios.PARENB)
+        raw.cflag |= (termios.CS8)
+        raw.lflag &= ~(termios.ICANON | termios.ECHO |
+                       termios.IEXTEN | (termios.ISIG * 1))
         raw.cc[termios.VMIN] = 1
         raw.cc[termios.VTIME] = 0
         tcsetattr(self.input_fd, termios.TCSADRAIN, raw)
@@ -367,7 +372,7 @@ class UnixConsole(Console):
         self.height, self.width = self.getheightwidth()
 
         self.__buffer = []
-        
+
         self.__posxy = 0, 0
         self.__gone_tall = 0
         self.__move = self.__move_short
@@ -375,40 +380,36 @@ class UnixConsole(Console):
 
         self.__maybe_write_code(self._smkx)
 
-        self.old_sigwinch = signal.signal(
-            signal.SIGWINCH, self.__sigwinch)
+        try:
+            self.old_sigwinch = signal.signal(
+                signal.SIGWINCH, self.__sigwinch)
+        except ValueError:
+            pass
 
     def restore(self):
         self.__maybe_write_code(self._rmkx)
         self.flushoutput()
         tcsetattr(self.input_fd, termios.TCSADRAIN, self.__svtermstate)
 
-        signal.signal(signal.SIGWINCH, self.old_sigwinch)
+        if hasattr(self, 'old_sigwinch'):
+            signal.signal(signal.SIGWINCH, self.old_sigwinch)
+            del self.old_sigwinch
 
     def __sigwinch(self, signum, frame):
         self.height, self.width = self.getheightwidth()
         self.event_queue.insert(Event('resize', None))
 
     def push_char(self, char):
-        self.partial_char += char
-        try:
-            c = unicode(self.partial_char, self.encoding)
-        except UnicodeError, e:
-            if len(e.args) > 4 and \
-                   e.args[4] == 'unexpected end of data':
-                pass
-            else:
-                raise
-        else:
-            self.partial_char = ''
-            self.event_queue.push(c)
-        
+        trace('push char {char!r}', char=char)
+        self.event_queue.push(char)
+
     def get_event(self, block=1):
         while self.event_queue.empty():
-            while 1: # All hail Unix!
+            while 1:
+                # All hail Unix!
                 try:
                     self.push_char(os.read(self.input_fd, 1))
-                except (IOError, OSError), err:
+                except (IOError, OSError) as err:
                     if err.errno == errno.EINTR:
                         if not self.event_queue.empty():
                             return self.event_queue.get()
@@ -460,7 +461,8 @@ class UnixConsole(Console):
             except KeyError:
                 height, width = struct.unpack(
                     "hhhh", ioctl(self.input_fd, TIOCGWINSZ, "\000"*8))[0:2]
-                if not height: return 25, 80
+                if not height:
+                    return 25, 80
                 return height, width
     else:
         def getheightwidth(self):
@@ -477,7 +479,7 @@ class UnixConsole(Console):
             if iscode:
                 self.__tputs(text)
             else:
-                os.write(self.output_fd, text.encode(self.encoding))
+                os.write(self.output_fd, text.encode(self.encoding, 'replace'))
         del self.__buffer[:]
 
     def __tputs(self, fmt, prog=delayprog):
@@ -527,10 +529,12 @@ class UnixConsole(Console):
                 e2 = self.event_queue.get()
                 e.data += e2.data
                 e.raw += e.raw
-                
+
             amount = struct.unpack(
                 "i", ioctl(self.input_fd, FIONREAD, "\0\0\0\0"))[0]
-            raw = unicode(os.read(self.input_fd, amount), self.encoding, 'replace')
+            data = os.read(self.input_fd, amount)
+            raw = unicode(data, self.encoding, 'replace')
+            #XXX: something is wrong here
             e.data += raw
             e.raw += raw
             return e
@@ -542,9 +546,11 @@ class UnixConsole(Console):
                 e2 = self.event_queue.get()
                 e.data += e2.data
                 e.raw += e.raw
-                
+
             amount = 10000
-            raw = unicode(os.read(self.input_fd, amount), self.encoding, 'replace')
+            data = os.read(self.input_fd, amount)
+            raw = unicode(data, self.encoding, 'replace')
+            #XXX: something is wrong here
             e.data += raw
             e.raw += raw
             return e
@@ -555,4 +561,3 @@ class UnixConsole(Console):
         self.__move = self.__move_tall
         self.__posxy = 0, 0
         self.screen = []
-
